#!/usr/bin/env python
# coding: utf-8


"""

======PROJECT======

MathEnergy

======DESCRIPTION======

The functions in this script are used to modify a Simbench dataset.

======Overview functions=====

01: def bus_correction(net)

02: def energy_pool_keep_mw(net, drop_fuels, keep_mw_fuel, new_slack_id)

03: def set_largest_slack_of_type(net, type, one_gen_only = True)

04: def distribution_factor_by_potential(net, profiles_abs, pp_id, 
                                         upper='max_p_mw', gen='gen')

05: def distribute_by_factor(energy_pool_ts, df_pp_distribution_factor_ts, 
                             df_pp_init_ts)

06: def clip_ts(net, df_pp_ts, pp_id, column)

07: def distribute_to_import(net, profiles_abs, energy_pool_ts, imp_share)

08: def distribute_to_offshore(net, profiles_abs, energy_pool_ts, repowering, 
                               new_off_share, ts_plot, equally)

09: def distribute_to_gas(net, profiles_abs, energy_pool_ts)

10: def distribute_max(net, profiles_abs, fuel, energy_pool_ts, 
                       upper = 'max_p_mw', gen = 'gen')

11: def gas_expansion_by_installed_capacity(net, profiles_abs, remain_ts, 
                                            gas_init_ts, share = 0.5, 
                                            plot_expansion = True, 
                                            plot_ts = True, geoplot = True, 
                                            ignore_new_powerplants = False)

12: def add_dc_line_no_sw(net)

13: def create_dcline_profiles(net, dc_capacity_dict, abs_profiles)

14: def conv_out_of_service_pp_to_gas(net, out_of_service_pp, type, 
                                      conversion_p_mw, energy_pool_ts, 
                                      profiles_abs, ignore_max_p_mw, 
                                      plot_ts = True, geoplot = True)

15: def distribute_energy_pool(net, repowering, energy_pool_ts, 
                               imp_share = 0.15, gas_max = True, 
                               plot_dist = False, plot_wind = False, 
                               plot_imp = False, repower_equally = True)

16: def increase_line_parallel_NEP(net)

17: def line_corrections(net)

======AUTHOR, CONTACT & DATE======

Paul Verwiebe, Department of Energy and Resource Management, Technical University of Berlin

verwiebe@tu-berlin.de

June 2021

"""



# ---IMPORT STATEMENTS---
import pandas as pd
import matplotlib.pyplot as plt

import pandapower as pp

from model.sb_modified import *
from model.draw import *
from model.tools import *

import logging
logger = logging.getLogger('mathenergy')



# ---FUNCTIONS---
# 01
def bus_corrections(net):
    """Performing a bus correction for topology adjustments.

    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.

    Returns
    -------
    None.

    """
    # Create new bus
    pp.create_bus(net, 
                  bus_name = 'Gronau', 
                  geodata = (8.059964, 53.326963),
                  vn_kv = 220)
   
    # Drop unconnected bus 
    # Shown after diagnostic tool is run
    net.bus.drop([3085], inplace = True)
    
    # Create new wind generation point
    pp.create_sgen(net, 
                   name = 'EHV Sgen ' + str(net.sgen.iloc[-1].name + 2),
                   p_mw = 0, 
                   bus = 1464,
                   type = 'wind offshore', 
                   in_service = True)
    
    net.sgen.loc[net.sgen.name == ('EHV Sgen ' + str(net.sgen.iloc[-1].name + 1)), 
                 'profile'] = 'WP2'


# 02
def energy_pool_keep_mw(net, drop_fuels, keep_mw_fuel, new_slack_id):
    """Removes selected power plant types (fuels) from network and saves generation timeseries in variable
    energy_pool_ts which can be distributed to other powerplants later

    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.
    
    drop_fuels : (list of str)
        Fuels to be dropped.
    
    keep_mw_fuel : (list of tuples: int, str)
        Generation capacity of power plant types (fuels) that are kept, while rest of the same fuel is dropped
    
    new_slack_id : (int)
        ID of the new slack.

    Returns
    -------
    energy_pool_ts : (pd.Series)
        Timeseries of remaining energy to be distributed.
    
    out_of_service_pp : (pd.DataFrame)
        Out-of-service coal power plants.
    
    profiles_abs : (dict of pd.DataFrame)
        Initial absolute profiles.
    
    net : (:obj:pandapowerNet)
        A pandapower grid.

    """
    # If fuels are kept, update net
    for keep_mw,fuel in keep_mw_fuel:
        a = net.gen.loc[net.gen.type == fuel].assign(new_max_p_mw = lambda x: ((x.max_p_mw / net.gen.loc[net.gen.type == fuel].max_p_mw.sum()) * keep_mw),
                                                     new_sn_mva = lambda x: ((x['sn_mva']/x['max_p_mw'])*x['new_max_p_mw']),
                                                     new_min_q_mvar = lambda x: ((x['min_q_mvar']/x['max_p_mw'])*x['new_max_p_mw']),
                                                     new_max_q_mvar = lambda x: ((x['max_q_mvar']/x['max_p_mw'])*x['new_max_p_mw']))
        net.gen.p_mw.update(a['new_max_p_mw'])
        # Set p_mw and max_pmw to the same level
        net.gen.max_p_mw.update(a['new_max_p_mw']) 
        net.gen.sn_mva.update(a['new_sn_mva'])
        net.gen.min_q_mvar.update(a['new_min_q_mvar'])
        net.gen.max_q_mvar.update(a['new_max_q_mvar'])
    
    # Get timeseries of absulute values
    profiles_abs = get_absolute_values(net)

    # Index of power plants to replace
    pp_replace = [index for index, row in net.gen.iterrows() if row.type in drop_fuels]

    # Sum of energy to replace
    energy_pool_ts = profiles_abs[('gen', 'p_mw')].loc[:, pp_replace].sum(axis = 1)

    # Find timesteps with negative balance
    profile_sums = compare_all_profiles(abs_profiles = profiles_abs, plot = False)
    neg_balance = profile_sums.total
    neg_balance.index = energy_pool_ts.index
    neg_balance[neg_balance > 0] = 0

    # Add negative balance to energy pool
    energy_pool_ts = energy_pool_ts - neg_balance

    # Edit net
    # Set power plants to be replaced out_of_service
    net.gen.loc[pp_replace, 'in_service'] = False

    # Set new slack
    net.ext_grid.loc[net.ext_grid.index != new_slack_id, 'in_service'] = False
	
	# Turn other nuclear power plants to generators and set service to FALSE
    pp.replace_ext_grid_by_gen(net,
                               ext_grids = net.ext_grid.loc[net.ext_grid.in_service == False].index,
                               cols_to_keep=net.ext_grid.columns)

    # Save out-of-service power plants
    out_of_service_pp = net.gen.loc[net.gen.in_service == False].sort_values(by = 'max_p_mw', 
                                                                             ascending = False)

    # Drop power plants that are being replaced
    pp.drop_out_of_service_elements(net)
    
    return energy_pool_ts, out_of_service_pp, profiles_abs, net


# 03
def set_largest_slack_of_type(net, type, one_gen_only = True):
    """Define the slack bus by capacity and fuel type.
    
    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.
        
    type : (str)
        Fuel type of slack.
        
    one_gen_only : (bool), optional
        Select only single generator buses. The default is TRUE.

    Returns
    -------
    None.

    """
    if one_gen_only:
        # The slack bus needs to be the only generator at the bus
        # All buses with multiple generators are discarded
        df_possible_slacks = pd.DataFrame(columns = net.gen.columns)
        for gen in net.gen.loc[net.gen.type == type].index:
            con_el = pp.get_connected_elements_	(net,buses=net.gen.loc[gen].bus, 
                                                 connected_other_elements = False)
            if len(con_el['gen']) == 1:
                df_possible_slacks = df_possible_slacks.append(net.gen.loc[gen])

        # Set largest generator of selection as slack
        slack_id = df_possible_slacks.max_p_mw.idxmax()
        pp.replace_gen_by_ext_grid(net, gens = slack_id, cols_to_keep = net.gen.columns)

    else:
        # Set largest power plant of selected type as ext-grid
        slack_id = net.gen.loc[net.gen.type == type, 'max_p_mw'].idxmax()
        pp.replace_gen_by_ext_grid(net, gens = slack_id, cols_to_keep = net.gen.columns)

    # Show all connected elements of ext_grid
    elements_of_slack = pp.get_connected_elements_dict(net, buses=net.ext_grid.bus,
                                                       connected_other_elements = False)
    msg = 'The following elements are connected to the slack bus {}: {}'.format(slack_id, 
                                                                                elements_of_slack)
    logger.info(msg)


# 04
def distribution_factor_by_potential(net, profiles_abs, pp_id, 
                                     upper = 'max_p_mw', gen = 'gen'):
    """Distribute energy_pool_ts by free potential distribution factor to 
    power plants or import points.

    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.

    profiles_abs : (dict of pd.DataFrame)
        Initial absolute profiles.

    pp_id : (list)
        IDs of power plants to distribute to.

    upper : (str), optional
        Column name of parameter to compute potential to.

    gen : (str), optional
        'gen' or 'sgen' depending on power plant to distribute to.

    Returns
    -------
    df_pp_distribution_factor_ts : (pd.DataFrame)
        Distribution factor of power plants per timestep.
        
    df_pp_init_ts : (pd.DataFrame)
        Initial timeseries of power plants.
        
    """
    # Initial timeseries of gas power plants
    df_pp_init_ts = profiles_abs[(gen, 'p_mw')].loc[:, pp_id]

    # Timeseries of free potential of gas power plants in MW
    df_pp_potential_ts = net.gen.loc[pp_id, upper] - df_pp_init_ts

    # Compute distribution factor per timestep by real free potential
    df_pp_distribution_factor_ts = df_pp_potential_ts.divide(net.gen.loc[pp_id, upper])
    df_pp_distribution_factor_ts = df_pp_distribution_factor_ts.divide(df_pp_distribution_factor_ts.sum(axis = 1), axis = 0)

    return df_pp_distribution_factor_ts, df_pp_init_ts


# 05
def distribute_by_factor(energy_pool_ts, df_pp_distribution_factor_ts, 
                         df_pp_init_ts):  
    """Distribute energy_pool_ts by free potential distribution factor to 
    power plants.

    Parameters
    ----------
    energy_pool_ts : (pd.Series)
        Timeseries of energy to be distributed.

    df_pp_distribution_factor_ts : (pd.DataFrame)
        Distribution matrix.

    df_pp_init_ts : (pd.DataFrame)
        Initial timeseries of power plants to distribute to.

    Returns
    -------
    df_pp_new_ts : (pd.DataFrame)
        New timeseries of power plants with distributed energy.
        
    """
    # Distributed energy
    df_pp_distributed_energy = df_pp_distribution_factor_ts.multiply((energy_pool_ts * 1), 
                                                                     axis = 'rows')
    
    # New timeseries of gas power plants with additional distributed energy
    df_pp_new_ts = df_pp_init_ts + df_pp_distributed_energy

    return df_pp_new_ts


# 06
def clip_ts(net, df_pp_ts, pp_id, column):
    """Clips values of timeseries which are distributed to a generator type according to a max value that cannot be exceded, e.g. max_p_mw.

    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.
        
    df_pp_ts : (pd.DataFrame)
        DataFrame with multiple timeseries to be clipped, indexed with an ID. 
		Energy pool to be distributed.
        
    pp_id : (string/list of strings)
        IDs of power plants to distribute to. 
        
    column : (str/list of str)
        property of power plants to define clip value (max value after which the timeseries is clipped)

    Returns
    -------
    df_pp_clipped_ts : (pd.DataFrame)
        generation timeseries of powerplants after clipping.
        
    energy_pool_ts : (pd.Series)
        Timeseries of remaining energy to be distributed.
        
    """
    # Clip every value in timeseries higher than max_p_mw (pp,t)
    temp = df_pp_ts.clip(upper = net.gen.loc[pp_id, column], axis = 1)

    # Difference between clipped and initial timeseries
    energy_pool_ts = (df_pp_ts - temp).sum(axis = 1)

    # df_new_gas_ts = clipped
    df_pp_clipped_ts = temp

    return df_pp_clipped_ts, energy_pool_ts


# 07
def distribute_to_import(net, profiles_abs, energy_pool_ts, imp_share):
    """Distribute energy_pool_ts to imports.
    
    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.
    
    profiles_abs : (dict of pd.DataFrame)
        Initial absolute profiles.
    
    energy_pool_ts : (pd.Series)
        Timeseries of energy to be distributed.
    
    imp_share : (float)
        Share of power distributed to imports (0 to 1).

    Raises
    ------
    ValueError

    Returns
    -------
    energy_pool_ts : (pd.Series)
        Timeseries of remaining energy to be distributed.
    
    df_import_init_ts : (pd.DataFrame) 
        Initial timeseries of imports. Can be used to plot before/after figures.
    
    df_new_import_ts : (pd.DataFrame) 
        Import timeseries after adjustments.
        
    """
    # Index of all import power plants imp0, imp1
    imports_id = [index for index, gen in net.gen.iterrows() if "imp" in gen.type]

    # Get distribution factor
    df_import_distribution_factor_ts, df_import_init_ts = distribution_factor_by_potential(net,
                                                                                           profiles_abs,
                                                                                           pp_id = imports_id,
                                                                                           upper = 'max_p_mw',
                                                                                           gen = 'gen')
    # Distribute by factor
    df_import_distributed_power = df_import_distribution_factor_ts.multiply((energy_pool_ts * imp_share), axis = 'rows')
	
    df_new_import_ts = df_import_distributed_power + df_import_init_ts
	
    if (df_new_import_ts > net.gen.loc[imports_id, 'max_p_mw'].values).any().any():
        df_new_import_ts, clipped_energy_pool_ts = clip_ts(net,
                                                           df_pp_ts = df_new_import_ts,
                                                           pp_id = imports_id,
                                                           column = 'max_p_mw')
        logger.info('Capacity of some import buses is reached. Remainings need to be distributed elsewhere.')

    # Remaining
    energy_pool_ts = energy_pool_ts - df_new_import_ts.sum(axis = 'columns') + df_import_init_ts.sum(axis = 'columns')

    if any(energy_pool_ts < 0):
        raise ValueError('Import share is to high. Boundaries are breached.')
		
    return energy_pool_ts, df_import_init_ts, df_new_import_ts


# 08
def distribute_to_offshore(net, profiles_abs, energy_pool_ts, repowering, 
                           new_off_share, ts_plot, equally):
    """Distribute energy_pool_ts to offshore wind.
    
    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.
   
    profiles_abs : (dict of pd.DataFrame)
        Initial absolute profiles.
    
    energy_pool_ts : (pd.Series)
        Timeseries of energy to be distributed.
    
    repowering : (int)
        Repower capacity for offshore wind in MW.
    
    new_off_share : (float)
        Share of repowering to new offshore wind parks.
    
    ts_plot : (bool)
		if True, plots initial and new offshore wind timeseries.

    equally : (bool)
		if True, offshore generation is equally distributed between existing offshore locations
		else generation is distributed proportional to installed capacities

    Returns
    -------
    energy_pool_ts : (pd.Series)
        Timeseries of remaining energy to be distributed.
    
    df_offshore_distributed_ts : (pd.DataFrame)
        timeseries of initial energy_pool_ts, which has been distributed to offshore wind
    
    df_offshore_init_ts : (pd.DataFrame)
        timeseries of initial offhsore generation
    
    energy_pool_neg : (pd.Series)
        Curtailment of offshore wind.

    """
    # IDs of all offshore wind power plants
    offshore_id = [index for index, row in net.sgen.iterrows() if row.type in ['wind offshore', 'Wind_Offshore']]

    # Timeseries of initial offshore wind capacity
    df_offshore_init_ts = profiles_abs[('sgen', 'p_mw')].loc[:, offshore_id]

    # Repowering old and new offshore wind parks by share
    if new_off_share>0:  

        logger.debug('Inland offshores available.')

        # Index of new_offshore wind parks with DC inland connection
        offshore_id = [index for index, row in net.sgen.iterrows() if row.type in 'Wind_Offshore']

        # Distribute equally
        new_capacity = new_off_share * repowering / len(offshore_id)

        # Set new offshore wind capacities
        net.sgen.loc[offshore_id, 'p_mw'] = new_capacity

        # Get offshore wind IDs of parks connected to coast
        offshore_id = [index for index, row in net.sgen.iterrows() if row.type in 'wind offshore']

        # Compute offshore wind distribution factor by initial capacity
        offshore_distribution_factor = net.sgen.loc[offshore_id, 'p_mw'] / net.sgen.loc[offshore_id, 'p_mw'].sum(axis = 'rows')

        # Share for new offshore windparks
        capacity_2_distribute = round(1 - new_off_share, 2) * repowering

        # Distribute __ GW by weighted installed power
        new_offshore_power = net.sgen.loc[offshore_id, 'p_mw'] + offshore_distribution_factor * capacity_2_distribute

        # Set new offshore wind capacities
        net.sgen.loc[offshore_id, 'p_mw'] = new_offshore_power

        logger.info('Distributed {} to coastal and {} to inland dc connected buses.'.format(round(1 - new_off_share, 2), new_off_share))

    # Only repowering of existing windparks
    else:  
	
        if equally:
		
            # Get offshore wind IDs of parks connected to coast
            offshore_id = [index for index, row in net.sgen.iterrows() if row.type in 'wind offshore']

            grouped = net.sgen.loc[offshore_id].groupby(by = 'bus').sum()
            grouped['count'] = net.sgen.loc[offshore_id].groupby(by = 'bus').count().name

            for bus in grouped.index:
                offshore_distribution_factor = grouped.loc[bus,'p_mw'] / grouped.loc[:,'p_mw'].sum(axis = 'rows')
                for el in offshore_id:
                    if net.sgen.loc[el,'bus'] == bus:
                        # Distribute repower
                        new_offshore_power = net.sgen.loc[el].p_mw + repowering / (len(grouped)*grouped.loc[bus, 'count'])

                        # Set new offshore wind capacities
                        net.sgen.loc[el, 'p_mw'] = new_offshore_power

                        # Set new offshore wind max_capacities
                        net.sgen.loc[el, 'max_p_mw'] = new_offshore_power			
			
        else:

            # Get offshore wind IDs of parks connected to coast
            offshore_id = [index for index, row in net.sgen.iterrows() if row.type in 'wind offshore']

            # Compute offshore wind distribution factor by initial capacity
            offshore_distribution_factor = net.sgen.loc[offshore_id, 'p_mw'] / net.sgen.loc[offshore_id, 'p_mw'].sum(axis = 'rows')
			
            # Distribute 2 GW by weighted installed power
            new_offshore_power = net.sgen.loc[offshore_id].p_mw + offshore_distribution_factor * repowering

            # Set new offshore wind capacities
            net.sgen.loc[offshore_id, 'p_mw'] = new_offshore_power

            # Set new offshore wind max_capacities
            net.sgen.loc[offshore_id, 'max_p_mw'] = new_offshore_power

    # Calculate new absolute profiles
    profiles_abs = sb.get_absolute_values(net, profiles_instead_of_study_cases = True)

    # IDs of both offshore wind types if available
    offshore_id = [index for index, row in net.sgen.iterrows() if row.type in ['wind offshore', 'Wind_Offshore']]

    # New offshore wind profiles
    df_offshore_new_ts = profiles_abs[('sgen', 'p_mw')].loc[:, offshore_id]

    # Difference of old and new offshore wind profiles
    df_offshore_distributed_ts = df_offshore_new_ts - df_offshore_init_ts

    # Remaining ts to distribute
    energy_pool_ts = energy_pool_ts.subtract(df_offshore_distributed_ts.sum(axis = 1))

    if any(energy_pool_ts < 0):
        # Replace negative values with zero
        energy_pool_neg = energy_pool_ts[energy_pool_ts < 0]
        energy_pool_ts[energy_pool_ts < 0] = 0

		
	# Plot initial and new offshore wind timeseries	
    if ts_plot:
        fig, ax = plt.subplots(figsize = (15,5))
        df_offshore_new_ts.sum(axis = 1).plot(ax = ax)
        df_offshore_init_ts.sum(axis = 1).plot(ax = ax)
        plt.ylabel('MW')
        plt.xlabel('Timestep')
        plt.legend(['New timeseries','Initial timeseries'])
        plt.title('Wind offshore repowering')
        plt.show()

    return energy_pool_ts, df_offshore_distributed_ts, df_offshore_init_ts, energy_pool_neg


# 09
def distribute_to_gas(net, profiles_abs, energy_pool_ts):
    """Distribute energy_pool_ts to gas.

    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.
    
    profiles_abs : (dict of pd.DataFrame)
        Initial absolute profiles.
    
    energy_pool_ts : (pd.Series)
        Timeseries of energy to be distributed.

    Returns
    -------
    energy_pool_ts : (pd.Series)
        Timeseries of remaining energy to be distributed.
    
    df_gas_distributed_ts : (pd.DataFrame)
        timeseries of initial energy_pool_ts, which has been distributed to gas pp 
    
    df_gas_init_ts : (pd.Series)
        Initial timeseries of gas generation.

    """
    # Index of all gas power plants
    gas_id = [index for index, row in net.gen.iterrows() if row.type in 'gas']

    # Get distribution factor
    df_gas_distribution_factor_ts, df_gas_init_ts = distribution_factor_by_potential(net, 
                                                                                     profiles_abs,
                                                                                     pp_id = gas_id,
                                                                                     upper = 'max_p_mw',
                                                                                     gen = 'gen')

    # Get new profiles
    df_gas_new_ts = distribute_by_factor(energy_pool_ts, df_gas_distribution_factor_ts, df_gas_init_ts)

    if (df_gas_new_ts > net.gen.loc[gas_id, 'max_p_mw'].values).any().any():

        df_gas_new_ts, energy_pool_ts = clip_ts(net,
                                                df_pp_ts = df_gas_new_ts,
                                                pp_id = gas_id,
                                                column = 'max_p_mw')

        logger.info('Capacity of some gas power plants is reached. Remainings need to be distributed elsewhere.')

    else:

        energy_pool_ts = None

        logger.info('Gas capacity was sufficient. Everything was distributed.')

    df_gas_distributed_ts = df_gas_new_ts - df_gas_init_ts

    logger.debug('Distribution to gas is done.')

    return energy_pool_ts, df_gas_distributed_ts, df_gas_init_ts


# 10
def distribute_max(net, profiles_abs, fuel, energy_pool_ts, upper = 'max_p_mw', 
                   gen = 'gen'):
    """Distribute an energy pool TS to powerplants up to maximum installed capacity by segmenting into exceeding and 
    non-exceeding timeseries.
    
    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.
    
    profiles_abs : (dict of pd.DataFrame)
        Initial absolute profiles.
    
    fuel : (list of/or str)
        Fuel type/s.
    
    energy_pool_ts : (pd.Series)
        Timeseries of energy to be distributed.
    
    upper : (str), optional
        Column name of parameter to compute potential to. 
        The default is 'max_p_mw'.
    
    gen : (str), optional
        'gen' or 'sgen' depending on power plant to distribute to. 
        The default is 'gen'.

    Returns
    -------
    energy_pool_ts : (pd.Series)
        Timeseries of remaining energy to be distributed.
    
    df_pp_distributed_ts : (pd.DataFrame)
        timeseries of initial energy_pool_ts, which has been distributed to power plants 
    
    df_pp_init_ts : (pd.DataFrame)
        Initial timeseries of power plants to distribute to.

    """
    # Index of all gas power plants
    pp_id = [index for index, row in net.gen.iterrows() if row.type in fuel]

    # Initial timeseries of gas power plants
    df_pp_init_ts = profiles_abs[(gen, 'p_mw')].loc[:, pp_id]

    # Timeseries of free potential of gas power plants in MW
    df_pp_potential_ts = net.gen.loc[pp_id, upper] - df_pp_init_ts

    # Timesteps exceeding upper value
    exceeding = df_pp_potential_ts.sum(axis = 1) < energy_pool_ts

    # IDs of not exceeding timesteps
    not_exceeding_id = [index for index, value in enumerate(exceeding.values) if not value]

    # Get distribution factor by free potential
    df_distribution_factor = df_pp_potential_ts.loc[not_exceeding_id, :].divide(
        df_pp_potential_ts.loc[not_exceeding_id, :].sum(axis = 1), axis = 'rows')

    # Initiate matrix of distributed energy
    df_pp_distributed_ts = pd.DataFrame(data = 0, index = df_pp_init_ts.index, 
                                        columns = df_pp_init_ts.columns)

    # Set values of distributed energy
    df_pp_distributed_ts.loc[not_exceeding_id, :] = df_distribution_factor.multiply(
        energy_pool_ts.loc[not_exceeding_id], axis = 'rows')

    # Clear not exceeding timesteps of energy pool as completely distributed
    energy_pool_ts.loc[not_exceeding_id] = 0

    # IDs of exceeding timesteps
    exceeding_id = [index for index, value in enumerate(exceeding.values) if value]

    # Set distributed energy of exceeding timesteps (difference to initial)
    df_pp_distributed_ts.loc[exceeding_id, :] = net.gen.loc[pp_id, upper].values - df_pp_init_ts.loc[exceeding_id, :]

    # Set remaining energy
    energy_pool_ts.loc[exceeding_id] = energy_pool_ts.loc[exceeding_id] - df_pp_distributed_ts.loc[exceeding_id, :].sum(
        axis='columns')

    return energy_pool_ts, df_pp_distributed_ts, df_pp_init_ts
	

# 11
def gas_expansion_by_installed_capacity(net, profiles_abs, remain_ts, 
                                        gas_init_ts, share = 0.5, 
                                        plot_expansion = True, plot_ts = True, 
                                        geoplot = True, 
                                        ignore_new_powerplants = False): 
    """Expand installed gas capacity to distribute remaining time series.
    
    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.
        
    profiles_abs : (dict)
        Initial absolute profiles.
    
    remain_ts : (pd.DataFrame)
        Power that needs to be distributed.
    
    gas_init_ts : (pd.DataFrame)
        Initial gas timeseries.
    
    share : (float), optional
        share of biggest (by installed capcity in MW) gas power plants are expanded. The default is 0.5.
    
    plot_expansion : (bool), optional
        Plots the initial and new gas timeseries. The default is TRUE.
        
    plot_ts : (bool), optional
        Plots the initial and new installed capacity per bus. The default is TRUE.
    
    geoplot : (bool), optional
        Map of locations of expanded gas power plants. The default is TRUE.
		
    ignore_new_powerplants : (bool), optional
        If TRUE, converted gas poweplants are not expanded. The default is FALSE.
        
    Returns
    -------
    profiles_abs : (dict of pd.DataFrame)
        Initial absolute profiles.
        
    p_mw_expan : (pd.DataFrame)
        Inital, new and overall installed capacity.
        
	"""
    if ignore_new_powerplants:
        pp_gas = net.gen.loc[net.gen.type == 'gas']
        pp_gas.drop(index = pp_gas.loc[pp_gas.name.str.contains('New')].index, inplace = True)
	
    # Filter for gas power plants
    else:
       pp_gas = net.gen.loc[net.gen.type == 'gas']

    # Get index of power plants with highest n percent of installed capacity
    pp_gas_highest = pp_gas.sort_values(by = 'max_p_mw', ascending = False).head(round(len(pp_gas) * share)).index
    max_p_mw = net.gen.loc[pp_gas_highest].max_p_mw

    # Get profile
    df_gas_ts = profiles_abs[('gen','p_mw')].loc[:,pp_gas_highest]

    # Remaining energy pool gets distributed
    factor = df_gas_ts.apply(lambda x: max_p_mw.div(max_p_mw.sum()), axis = 1)
    df_diff_gas_ts = factor.T.multiply(remain_ts).T
	
    # Compute new timeseries
    df_new_gas_ts = df_gas_ts + df_diff_gas_ts
    profiles_abs[('gen','p_mw')].loc[:,pp_gas_highest] = df_new_gas_ts
    
    # Compute the expansion
    net.gen.loc[pp_gas.index,'p_mw_expan'] = df_new_gas_ts.max(axis = 0).sub(net.gen.loc[pp_gas.index].max_p_mw)
    p_mw_expan = net.gen.loc[pp_gas.index][['max_p_mw','p_mw_expan']]
    p_mw_expan = p_mw_expan.assign(new_max_p_mw = p_mw_expan.sum(axis = 1))
    
    # Plot expansion
    if plot_expansion:
        max_p_mw_expan_grouped = net.gen.loc[pp_gas.index].groupby(by = 'bus').sum()[['max_p_mw', 'p_mw_expan']]
        max_p_mw_expan_grouped = max_p_mw_expan_grouped.assign(new_maxp_mw = max_p_mw_expan_grouped.sum(axis = 1))
        max_p_mw_expan_grouped = max_p_mw_expan_grouped.sort_values(by = (['new_maxp_mw']))
        ax = max_p_mw_expan_grouped[['max_p_mw','p_mw_expan']].plot.bar(figsize = (15,5), stacked = True, ylim = (0,6000))
        ax.legend(['Initial installed capacity','Expansion'])
        ax.set_ylabel('MW')
        plt.show()
    
    # Plot timeseries
    if plot_ts:
        fig,ax = plt.subplots(figsize = (15,5))
        df_new_gas_ts.sum(axis = 1).plot(ax = ax)
        df_gas_ts.sum(axis = 1).plot(ax = ax)
        gas_init_ts.sum(axis = 1).plot(ax = ax)
        plt.ylabel('MW')
        plt.xlabel('Timestep')
        plt.legend(['After expansion','After distribution utilizing full initial installed capacity', 'Initial before distribution'])
        plt.show()
		
    if geoplot:
        lc = pplot.create_line_collection(net, 
                                          net.line.index, 
                                          color = 'lightgrey', 
                                          use_bus_geodata = True, 
                                          zorder = 1)
        
        bc = pplot.create_bus_collection(net, 
                                         net.bus.index, 
                                         zorder = 2, 
                                         color = 'lightgrey', 
                                         size = 0.02)
		
        loadc = pplot.create_bus_collection(net,
                                            buses = net.gen.loc[net.gen.p_mw_expan >= 0].bus,
                                            z = net.gen.loc[net.gen.p_mw_expan >= 0].p_mw_expan,
                                            size = 0.06,
                                            cmap = 'viridis_r',
                                            cbar_title = 'Gas capacity expansion [MW]',
                                            zorder = 3)

        pplot.draw_collections([lc, bc, loadc])
        
    return profiles_abs, p_mw_expan
					 

# 12
def add_dc_line_no_sw(net):
    """Rename existing DC lines and add new DC lines to the net according to 
    NEP 2019.

    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.
        
    Returns
    -------
    None.
    
    """
    # Rename existing lines
    net.dcline.name = ['P65', 'DC2', 'DC5', 'DC1', 'DC4', 'DC3']
    # Set transfer capacity
    net.dcline.max_p_mw = [1000, 2000, 2000, 2000, 4000, 2000]
    
    # Add missing DC lines from NEP
    dc_nep = {'DC21b':2000, 'DC25':2000, 'DC20':2000}

    # DC21a: Heide/West - Wilhelmshaven 2 not confirmed in NEP 2030
    # DC23: Uentrop - Altbach not confirmed in NEP 2030
    # 1476 = Substation Maade = Wilhelmshaven 2
    # 158 = Station Uentrop
    # 412 = EnBW substation Altbach 1
    """DC21b: Wilhelmshaven 2 - Uentrop"""                                                                           
    pp.create_dcline(net,
                     name = 'DC21b',
                     p_mw = 0, 
                     loss_mw = 0,
                     from_bus = 1476, 
                     to_bus = 158, 
                     loss_percent = 1.2, 
                     vm_from_pu = 1.092, 
                     vm_to_pu = 1.092,
                     max_p_mw = dc_nep['DC21b'],
                     in_service = True)
    
    """DC25: Heide / West – Polsum"""
    # Heide/West does not exist! 30 = 1460 = Substation nuclear power plant Brunsbüttel
    # 30 is on 380 kV and 1460 on 220 kV
    # Since Polsum is on 220 kV, select 1460 as the connection node.
    # 1572 = Station Bergmannsglück, 1526 = Kusenhorst
    # Polsum is located at 51.625148, 7.022988 --> 1572 corresponds approximately to Polsum
    # EHV Gen 65 has a vm_pu of 1.068
    pp.create_dcline(net, 
                     name = 'DC25',
                     p_mw = 0, 
                     loss_mw = 0,
                     from_bus = 1460, 
                     to_bus = 1572, 
                     loss_percent = 1.2, 
                     vm_from_pu = 1.068, 
                     vm_to_pu = 1.068,
                     max_p_mw = dc_nep['DC25'],
                     in_service = True)

    """DC20: Klein Rogahn - Isar"""
    # Klein-Rogahn corresponds to Görries accoridng to ENTSO-E map
    # 142 = Görries, 140 = Wessin, 376 = Isar
    # EHV Gen 266 at bus 142 has a vm_pu of 1.092                                                                
    pp.create_dcline(net, 
                     name = 'DC20',
                     p_mw = 0, 
                     loss_mw = 0,
                     from_bus = 142, 
                     to_bus = 48, 
                     loss_percent = 1.2, 
                     vm_from_pu = 1.092, 
                     vm_to_pu = 1.092,
                     max_p_mw = dc_nep['DC20'],
                     in_service = True)
    
    # Change direction of DC 1 to go from north to south
    net.dcline.loc[net.dcline.name == 'DC1', 'from_bus'] = 1738
    net.dcline.loc[net.dcline.name == 'DC1', 'to_bus'] = 1564


# 13 
def create_dcline_profiles(net, dc_capacity_dict, abs_profiles):    
    """Create new profiles for DC lines.

    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.

    dc_capacity_dict : (dict)
        key: 'name of the dcline', tuples of: 'max_p_mw', 'connecting dcline' 
        and 'nearest wind gen bus'.

    abs_profiles : (dict)
        Absolute profiles.

    Returns
    -------
    abs_profiles : (dict)
        New absolute profiles.
        
    """
    # Initiate matrix of distributed energy
    df_dc_profile = pd.DataFrame(data = 0, index = abs_profiles[('sgen', 'p_mw')].index, columns = net.dcline.index)   

    for line in dc_capacity_dict.keys():
        # Find DC line ID in net object
        dc_id = net.dcline.loc[net.dcline.name == line].index

        # For DC line with wind gen starting point
        if dc_capacity_dict[line][2] != None:
            # Find wind ID in absolute profiles
            wind_id = net.sgen.loc[net.sgen.type == 'wind offshore'].loc[net.sgen.bus == dc_capacity_dict[line][2]].index
            # Set wind profile as DC line profile
            df_dc_profile[dc_id] = abs_profiles[('sgen', 'p_mw')].loc[:,wind_id].sum(axis = 1)
            # Timesteps exceeding upper value
            exceeding = dc_capacity_dict[line][0] < abs_profiles[('sgen', 'p_mw')].loc[:,wind_id].sum(axis = 1)
            # For exceeding timesteps set max_p_mw
            df_dc_profile.loc[exceeding, dc_id] = dc_capacity_dict[line][0]
			
        net.dcline.loc[net.dcline.name == line, 'max_p_mw'] = dc_capacity_dict[line][0]

    for line in dc_capacity_dict.keys():
        # Find DC line ID in net object
        dc_id = net.dcline.loc[net.dcline.name == line].index
		
        # For DC line with feed-in from connected DC line
        if dc_capacity_dict[line][1] != None:
            connected_dcline = dc_capacity_dict[line][1]
            # For profile add connecting DC line feed-in to already excisting wind power generation
            df_dc_profile[dc_id] = (df_dc_profile[dc_id].iloc[:,0] + df_dc_profile[net.dcline.loc[net.dcline.name == connected_dcline].index].iloc[:,0]).to_frame()
            # Timesteps exceeding upper value
            exceeding = df_dc_profile[dc_id] > dc_capacity_dict[line][0]
            exceeding = exceeding.iloc[:,0]
            # For exceeding timesteps set max_p_mw
            df_dc_profile.loc[exceeding, dc_id] = dc_capacity_dict[line][0]

    # Save new profiles		
    abs_profiles[('dcline', 'p_mw')] = df_dc_profile
    
    return abs_profiles
	

# 14
def conv_out_of_service_pp_to_gas(net, out_of_service_pp, type, conversion_p_mw, 
                                  energy_pool_ts, profiles_abs, ignore_max_p_mw, 
                                  plot_ts = True, geoplot = True):    
    """Convert out-of-service poweplants to gas power plants.

    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.

    out_of_service_pp : (pd.DataFrame)
        Out-of-service power plants.
    
    type : (list)
        Fuel types to be converted.

    conversion_p_mw : (int)
        Power of power plants to be converted.

    energy_pool_ts : (pd.DataFrame)
        Power that needs to be distributed.

    profiles_abs : (dict)
        Absolute profiles.
    
    ignore_max_p_mw : (bool)
        If TRUE, max_p_mw expansion until energy pool is 0.
    
    plot_ts : (bool), optional
        Plots the timeseries. The default is TRUE.
		
    geoplot : (bool), optional
        Plots the converted power plants. The default is TRUE.

    Returns
    -------
    profiles_abs : (dict)
        New absolute profiles.
    
    energy_pool_ts : (pd.DataFrame)
        Power that needs to be distributed.
        
    p_mw_expan : (pd.DataFrame)
        Inital, new and overall installed capacity.

    Notes
    -----
    - :obj: net is modified for offshore wind capacities.
    
    """
    # Filter for potential power plants to convert
    out_of_service_pp = out_of_service_pp[out_of_service_pp.type.isin(type)]
    out_of_service_pp = out_of_service_pp.sort_values(by = 'max_p_mw', ascending = False)
    conv_pp = pd.DataFrame(columns=out_of_service_pp.columns)
    new_gas_p_mw = 0
    counter = 1
    
    # Convert power plants to gas power plants until conversion_p_mw is reached
    while new_gas_p_mw < conversion_p_mw:
        out_of_service_pp.loc[out_of_service_pp.iloc[0].name,['name', 'type', 'profile']] = np.array(['New_gas_' + str(counter), 'gas', 'pp_new' + str(counter)])
        conv_pp = conv_pp.append(other=out_of_service_pp.iloc[0,:])
        out_of_service_pp.drop(out_of_service_pp.head(1).index, axis = 0, inplace = True)
        new_gas_p_mw = conv_pp.max_p_mw.sum()
        counter += 1
    
    # Append new power plants
    for el in conv_pp.index:
        pp.create_gen(net,
                      index = el,
                      bus = conv_pp.loc[el, 'bus'],
                      p_mw = conv_pp.loc[el, 'p_mw'],
                      vm_pu = conv_pp.loc[el, 'vm_pu'],
                      name = conv_pp.loc[el, 'name'],
                      type = conv_pp.loc[el, 'type'],
                      max_p_mw = conv_pp.loc[el, 'max_p_mw'],
                      max_q_mvar = conv_pp.loc[el, 'max_q_mvar'],
                      min_q_mvar = conv_pp.loc[el, 'min_q_mvar'],
                      sn_mva = conv_pp.loc[el, 'sn_mva'])
    
    # Compute distribution factor by max p_mw
    distribution_factor = net.gen.loc[conv_pp.index, 'max_p_mw'] / net.gen.loc[conv_pp.index, 'max_p_mw'].sum(axis = 'rows')
	
    # Timesteps exceeding upper value
    exceeding = conv_pp.loc[:,'max_p_mw'].sum(axis = 'rows') < energy_pool_ts

    # Initiate matrix of distributed energy
    df_pp_distributed_ts = pd.DataFrame(data = 0, index = energy_pool_ts.index, columns = conv_pp.index)

    # Compute profiles
    for col in df_pp_distributed_ts:
        df_pp_distributed_ts[col] = energy_pool_ts
        # For non-exceeding timesteps propotional to max_p_mw
        df_pp_distributed_ts.loc[~exceeding, col] = df_pp_distributed_ts.loc[~exceeding, col].multiply(distribution_factor.loc[col])
        # For exceeding timesteps set max_p_mw
        df_pp_distributed_ts.loc[exceeding, col] = conv_pp.loc[col, 'max_p_mw']
        profiles_abs[('gen', 'p_mw')][col] = df_pp_distributed_ts[col]

    # Compute remaining energy pool
    energy_pool_ts_old = energy_pool_ts.copy()
    energy_pool_ts.loc[~exceeding] = 0
    energy_pool_ts.loc[exceeding] = energy_pool_ts.loc[exceeding] - df_pp_distributed_ts.loc[exceeding].sum(axis = 1)
	
    if plot_ts:
        fig,ax = plt.subplots(figsize = (15,5))
        energy_pool_ts_old.plot(ax = ax)
        energy_pool_ts.plot(ax = ax)
        plt.ylabel('MW')
        plt.xlabel('Timestep')
        plt.legend(['Before conversion', 'Remaining energy pool'])
        plt.show()
		
    if geoplot:
        lc = pplot.create_line_collection(net, 
                                          net.line.index, 
                                          color = 'lightgrey', 
                                          use_bus_geodata = True, 
                                          zorder = 1)
        
        bc = pplot.create_bus_collection(net, 
                                         net.bus.index, 
                                         zorder = 2, 
                                         color = 'lightgrey', 
                                         size = 0.02)

        convc = pplot.create_bus_collection(net,
                                            buses = net.gen.loc[net.gen.name.str.startswith('New_gas', na = False)].bus,
                                            z = net.gen.loc[net.gen.name.str.startswith('New_gas', na = False)].p_mw,
                                            size = 0.06,
                                            cmap = 'viridis_r',
                                            cbar_title = 'Gas expansion [MW]',
                                            zorder = 3)

        pplot.draw_collections([lc, bc, convc])
		
    if ignore_max_p_mw:
        
        # Filter for gas power plants
        expan_pp = net.gen.loc[net.gen.name.str.contains('New')]

        # Get index of power plants with highest n percent of installed capacity
        # pp_gas_highest = pp_gas.sort_values(by = 'max_p_mw', ascending = False).head(round(len(pp_gas) * share)).index
        max_p_mw = expan_pp.max_p_mw

        # Get profile
        df_gas_ts = profiles_abs[('gen', 'p_mw')].loc[:,expan_pp.index]

        # Remaining energy pool gets distributed
        factor = df_gas_ts.apply(lambda x: max_p_mw.div(max_p_mw.sum()), axis = 1)
        df_diff_gas_ts = factor.T.multiply(energy_pool_ts).T
		
        # Compute new timeseries
        df_new_gas_ts = df_gas_ts + df_diff_gas_ts
        profiles_abs[('gen', 'p_mw')].loc[:,expan_pp.index] = df_new_gas_ts
    
        # Compute the expansion
        net.gen.loc[expan_pp.index, 'p_mw_expan'] = df_new_gas_ts.max(axis = 0).sub(net.gen.loc[expan_pp.index].max_p_mw)
        p_mw_expan = net.gen.loc[expan_pp.index][['max_p_mw', 'p_mw_expan']]
        p_mw_expan = p_mw_expan.assign(new_max_p_mw=p_mw_expan.sum(axis = 1))
		
        max_p_mw_expan_grouped = net.gen.loc[expan_pp.index].groupby(by = 'bus').sum()[['max_p_mw', 'p_mw_expan']]
        max_p_mw_expan_grouped = max_p_mw_expan_grouped.assign(new_maxp_mw = max_p_mw_expan_grouped.sum(axis = 1))
        max_p_mw_expan_grouped = max_p_mw_expan_grouped.sort_values(by = (['new_maxp_mw']))
        ax = max_p_mw_expan_grouped[['max_p_mw', 'p_mw_expan']].plot.bar(figsize = (15,5), stacked = True, ylim = (0,6000))
        ax.legend(['Initial installed capacity', 'Expansion'])
        ax.set_ylabel('MW')
        plt.show()
		
        fig,ax = plt.subplots(figsize = (15,5))
        df_new_gas_ts.sum(axis = 1).plot(ax = ax)
        df_gas_ts.sum(axis = 1).plot(ax = ax)
        plt.ylabel('MW')
        plt.xlabel('Timestep')
        plt.legend(['After expansion', 'After conversion utilizing full initial installed capacity'])
        plt.show()
		
    else:
        p_mw_expan = pd.DataFrame()
		
    return energy_pool_ts, profiles_abs, p_mw_expan
	

# 15
def distribute_energy_pool(net, repowering, energy_pool_ts, imp_share = 0.15, 
                           gas_max = True, plot_dist = False, plot_wind = False, 
                           plot_imp = False, repower_equally = True):
    """Distribute energy of shut-off powerplants (energy_pool_ts) to imports, offshore wind
	and gas power plants. For imports and gas power plants installed capacity is respected.
	For offshore wind, additional generation capacity is built.

    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.

    imp_share : (float), optional
        Share of power to imports (0 to 1). The default is 0.15.

    repowering : (int)
        Repower capacity for offshore wind in MW.

    energy_pool_ts : (pd.DataFrame)
        Power that needs to be distributed.

    gas_max : (bool), optional
        If TRUE, energy will be distributed to gas until max_p_mw is reached.
        If FALSE, energy will be distributed by potential and clipped above max_p_mw.
        The default is TRUE.

    plot_dist : (bool), optional
        Plots the distribution graph. The default is FALSE.
    
    plot_wind : (bool), optional
        Plots initial and new offshore wind timeseries. The default is FALSE.
    
    plot_imp : (bool), optional
        Plots initial and new import timeseries. The default is FALSE.

    repower_equally : (bool), optional
        The default is TRUE.

    Returns
    -------
    profiles_abs : (dict)
        New absolute profiles.
        
    energy_pool_ts : (pd.Series)
        Energy that needs to be distributed.
        
    df_gas_init_ts : (pd.Series)
        Initial timeseries of gas generation.
        
    energy_pool_neg : (pd.Series)
        Curtailment of offshore wind.

    Notes
    -----
    - :obj: net is modified for offshore wind capacities.
    - Initial profiles are derived from (:obj: net).
    - Distribution to imports:
        - X% (default imp_share = 15%) of power is distributed to import coupling points.
        - The share of power exceeds the minimum potential of the import 
          coupling points.
        - The distribution is weighted by the minimum potential of the import 
          coupling points.
        
    """
    # Calculate new absolute profiles
    profiles_abs = sb.get_absolute_values(net, profiles_instead_of_study_cases = True)
    
    
    ###############
    # OFFSHORE WIND
    ###############

    energy_pool_ts, df_offshore_distributed_ts, df_offshore_init_ts, energy_pool_neg = distribute_to_offshore(net = net,
                                                                                                              profiles_abs = profiles_abs,
                                                                                                              energy_pool_ts = energy_pool_ts,
                                                                                                              repowering = repowering,
                                                                                                              new_off_share = 0,
                                                                                                              ts_plot = plot_wind,
                                                                                                              equally = repower_equally)

    logger.debug('Distribution to offshore wind is done.')


    ###############
    # IMPORTS
    ###############

    energy_pool_ts, df_import_init_ts, df_new_import_ts  = distribute_to_import(net = net,
                                                                                profiles_abs = profiles_abs,
                                                                                energy_pool_ts = energy_pool_ts,
                                                                                imp_share = imp_share)

    logger.debug('Distribution to imports is done.')
    
    # Plot initial and new import timeseries	
    if plot_imp:
        fig, ax = plt.subplots(figsize = (15,5))
        df_new_import_ts.sum(axis = 1).plot(ax = ax)
        df_import_init_ts.sum(axis = 1).plot(ax = ax)
        plt.ylabel('MW')
        plt.xlabel('Timestep')
        plt.legend(['New timeseries', 'Initial timeseries'])
        plt.title('Import timeseries')
        plt.show()


    ###############
    # GAS
    ###############

    # Distribute until max capacity is reached
    if gas_max:
        energy_pool_ts, df_gas_distributed_ts, df_gas_init_ts = distribute_max(net = net,
                                                                               profiles_abs = profiles_abs,
                                                                               fuel = 'gas',
                                                                               energy_pool_ts = energy_pool_ts,
                                                                               upper = 'max_p_mw',
                                                                               gen = 'gen')
    
    # Distribute by relative potential and then clip
    else:
        energy_pool_ts, df_gas_distributed_ts, df_gas_init_ts = distribute_to_gas(net = net,
                                                                                  profiles_abs = profiles_abs,
                                                                                  energy_pool_ts = energy_pool_ts)

    logger.debug('Distribution to gas is done.')


    ###############
    # NEW PROFILES
    ###############

    # ## OFFSHORE WIND
    # Calculate new absolute profiles
    profiles_abs = sb.get_absolute_values(net, profiles_instead_of_study_cases = True)
    
    # Wind curtailment			
    wind_pp = net.sgen.loc[net.sgen.type == 'wind offshore'].index
    factor = profiles_abs[('sgen', 'p_mw')].loc[energy_pool_neg.index,wind_pp].div(profiles_abs[('sgen', 'p_mw')].loc[energy_pool_neg.index, wind_pp].sum(axis = 1), axis = 0)
    profiles_abs[('sgen', 'p_mw')].loc[energy_pool_neg.index,wind_pp] = profiles_abs[('sgen', 'p_mw')].loc[energy_pool_neg.index, wind_pp].add(factor.mul(energy_pool_neg, axis = 0))

    # ## GAS
    # Get gas IDs again
    gas_id = [index for index, row in net.gen.iterrows() if row.type in 'gas']

    # Dub new gas profiles
    profiles_abs[('gen', 'p_mw')].loc[:, gas_id] = df_gas_init_ts + df_gas_distributed_ts

    # ## IMPORTS
    # Index of all import power plants imp0, imp1
    imports_id = [index for index, gen in net.gen.iterrows() if "imp" in gen.type]

    # Add share to absolute profiles of imports
    profiles_abs[('gen', 'p_mw')].loc[:, imports_id] = df_new_import_ts
	
    # Plot distribution
    if plot_dist:
        plt.subplots(figsize = (15,5))
        plt.stackplot(df_offshore_distributed_ts.index,
                      [(df_new_import_ts-df_import_init_ts).sum(axis = 'columns').values.T,
                      df_offshore_distributed_ts.sum(axis = 'columns').values.T,
                      df_gas_distributed_ts.sum(axis = 'columns').values.T,
                      energy_pool_ts])
        plt.ylabel('MW')
        plt.xlabel('Timestep')
        plt.legend(['Import', 'Wind offshore', 'Gas', 'Remaining'])
        plt.title('Distribution of energy pool')
        plt.show()

    return profiles_abs, energy_pool_ts, df_gas_init_ts, energy_pool_neg
	

# 16
def increase_line_parallel_NEP(net):	
    """Add parallel lines.
    
    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.

    Returns
    -------
    None.
    
    """
    # P215
    net.line.loc[289, 'parallel'] = 3
    
    # DC line connection
    net.line.loc[765, 'parallel'] = 2
    
    # TTG-P67
    net.line.loc[100, 'parallel'] = 3

    # TTG-P66
    net.line.loc[747, 'parallel'] = 2

    # TTG-P66 (parallel to 747)
    net.line.loc[746, 'parallel'] = 2

    # TTG-P66
    net.line.loc[692, 'parallel'] = 2

    # TTG-P66 (parallel to 691)
    net.line.loc[691, 'parallel'] = 2

    # NOR-2-2
    net.line.loc[593, 'parallel'] = 2

    # AMP-009
    net.line.loc[79, 'parallel'] = 2
    
    # AMP-009
    net.line.loc[768, 'parallel'] = 2
	

# 17
def line_corrections(net):
    """Correction of faulty or unrealistic properties of certain lines.
    
    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.

    Returns
    -------
    None.
    
    """
    # ## EHV line 825
    # Only generators (no loads) are connected to bus 1967. Hence EHV line 825, 
    # 820 and 824 should have appropriate capacity to transport the energy of 
    # generators at bus 1967 to other buses. 
    # Kraftwerk Neurath 4GW installed power!
    net.line.loc[824, 'parallel'] = 4
    
    # ## EHV line 769
    # At the end of line EHV 769, at bus 1536, 3173 MW gen power installed while 
    # the line has a transport capacity of only 495 MW. Since increasing the 
    # number of parallel lines by factor 6 would satisfy the transport demand 
    # the line type is changed to have a higher max_i_ka of 2.6. 
    # Kernkraftwerk Emsland!
    # Change line type
    net.line.std_type.loc[[768]] = 'LineType_2'
    # Change line parameters according to line type
    net.line.c_nf_per_km.loc[[768]] = net.std_types['line']['LineType_2']['c_nf_per_km']
    net.line.r_ohm_per_km.loc[[768]] = net.std_types['line']['LineType_2']['r_ohm_per_km']
    net.line.x_ohm_per_km.loc[[768]] = net.std_types['line']['LineType_2']['x_ohm_per_km']
    net.line.max_i_ka.loc[[768]] = net.std_types['line']['LineType_2']['max_i_ka']
    # Increase number of parallel lines
    net.line.loc[768, 'parallel'] = 4
    
    # ## EHV line 367
    # Node 1450 = Umspannwerk Neufinsing east of Munich. The 220 kV line 
    # Neufinsing Irsching is parallel to the 380 kV line Ottenhofen Irsching. In 
    # reality the TSO will somehow manage the flow between the two lines but the 
    # DC load flow is not capable of doing so. Node 1830 = Irsching. EHV line 107 
    # is at 220 kV but in reality it is a 380 kV line! Node 1452 should be at 
    # 380 kV but it is actually at 220 kV. As some kind of remedial action one 
    # could set a parallel 220 kV line to EHV line 107.
    # net.line.loc[106,'parallel'] = 2 --> See explanation above
    # Neufinsing-Zolling (EHV 367) has capacity of 2.2 kA. Will be adjusted in 
    # the network. Zolling-Irsching has only 1.6 kA (EHV 306).
    net.line.loc[366, 'max_i_ka'] = 2.2
    net.line.loc[305, 'max_i_ka'] = 1.6
    # EHV 107 Ottenhofen-Irsching has 380 kV and 2.2 kA in reality, but here 
    # only 1.3 kA and 220 kV. Set line type to 2 to achieve at least a higher capacity.
    pp.change_std_type(net, eid = 106, element = 'line', name = 'LineType_2')
    
    # ## EHV line 809
    # This line is inside of a substation! It connects two parts of this 
    # substation (see Google Maps). It is unlikely that a line inside a substation 
    # can be congested, because it would mean that the substation would be 
    # incorrectly designed. Solution: Change the line type to have twice the capacity!
    net.line.std_type.loc[[808]] = 'LineType_1'
    # Change line parameters according to line type
    net.line.c_nf_per_km.loc[[808]] = net.std_types['line']['LineType_1']['c_nf_per_km']
    net.line.r_ohm_per_km.loc[[808]] = net.std_types['line']['LineType_1']['r_ohm_per_km']
    net.line.x_ohm_per_km.loc[[808]] = net.std_types['line']['LineType_1']['x_ohm_per_km']
    net.line.max_i_ka.loc[[808]] = net.std_types['line']['LineType_1']['max_i_ka']
    
    # ## EHV Line 15 (14) 
    # Herbertingen-Bürs: Connects import generator with Germany. Generator has 
    # 1370 MW installed power, but line only provides 250 MW capacity.
    # 1728 = Herbertingen, 1726 = Bürs, 428 = Grünkraut, 290 = Obermooweiler. 
    # Net topology is wrong (see ENTSOE grid map). There is a Grünkraut Herbertingen 
    # (1728 to 428) connection with 1 GW capacity and 380 kV. Grünkraut Obermooweiler 
    # has 1 GW. (428 to 290) = EHV 537. Replace EHV 15 with 380 kV line.
    pp.drop_lines(net, [14])
    new_line_ind = pp.create_line(net, from_bus = 294, to_bus = 428, 
                                  length_km = 43.1, std_type = 'LineType_4', 
                                  name = None, index = 14, geodata = None, 
                                  df = 1.0, parallel = 1, in_service = True, 
                                  max_loading_percent = 100)
    net.line.loc[new_line_ind, 'name'] = 'EHV Line ' + str(new_line_ind+1)
    net.line.loc[new_line_ind, 'subnet'] = 'EHV1'
    net.line.loc[new_line_ind, 'voltLvl'] = 1
    # Set same line type for EHV 537 and EHV 98
    pp.change_std_type(net, eid = 536, element = 'line', name = 'LineType_4')
    pp.change_std_type(net, eid = 97, element = 'line', name = 'LineType_4')
    
    # ## EHV line 410
    # Looks like north to south bottleneck. No obvious error.
    
    # ## EHV line 615 (614) 
    # 1558 = Rommerskirchen, 1568 = Brauweiler.
    # Has capacity of 3.1 kA according to Amprion. Change line type --> 2.6 kA
    net.line.loc[614, 'std_type'] = 'LineType_2'
    net.line.loc[614, 'c_nf_per_km'] = net.std_types['line']['LineType_2']['c_nf_per_km']
    net.line.loc[614, 'r_ohm_per_km'] = net.std_types['line']['LineType_2']['r_ohm_per_km']
    net.line.loc[614, 'x_ohm_per_km'] = net.std_types['line']['LineType_2']['x_ohm_per_km']
    net.line.loc[614, 'max_i_ka'] = net.std_types['line']['LineType_2']['max_i_ka']
    
    # ## EHV line 170 (169)
    # Dillinger Hütte is connected to the line. The line capacity is 247 MW, while 
    # at buses 1852, 1854 and 1856 there are loads connected with a total installed 
    # power of 313 MW. They can only be supplied by EHV line 170. No 
    # other generators are connected to these buses. Hence this is an artificial 
    # bottleneck since in reality lower grid levels would support the supply of 
    # these loads. Line 170 only serves as a transport line to supply the loads at 
    # their end without further effects on the transmission grid and other nodes.
    # This artificial bottleneck should therefore be remedied. Another reason of 
    # this problem could be the wrong assignment of loads to buses. The load of 
    # 313 MW was not part of the initial simbench dataset but of the present one,
	# although loads were untouched.
	# In Simbench there are 180 MW installed at the three nodes in total!
    net.line.loc[169, 'parallel'] = 2
    
    # ## EHV line 666 (665)
    # EHV 666 only supplies the load at node 1660, which has an installed power 
    # of 774 MW. Since the line can only enable a flow of 495 MW, it is doubled
    # in order to provide twice the capacity. This is done by adding a parallel 
	# line. At node 1650 there should be some load,
    # since a ThyssenKrupp steel mill (Beeckerwerth) is connected. Perhaps it was 
    # aggregated with the load of the Arcelor Mittal steel mill connected to 
    # node 1660. Furthermore, the Heizkraftwerk Hamborn, which has an installed 
    # capacity of 240 MW, is missing at node 1650 (https://de.wikipedia.org/wiki/Heizkraftwerk_Hamborn).
    net.line.loc[665, 'parallel'] = 2

	# ## EHV line 755 (754)
    # At node 430 (end of line EHV 755) there is the nuclear plant Grundremmingen 
    # that is only connected to EHV line 755. Hence the capacity of the line must 
    # be choosen accordingly to the installed power of the plant.
    net.line.loc[754, 'parallel'] = 2


	# ## EHV line 557 (556)
    # Connects nuclear power plant Neckarwestheim to the transmission grid, hence 
    # its capacity should be choosen accordingly to the installed power at node 326,
	# where the pp is located. EHV line 802 connects the nuclear power plant with a substation.
	# EHV line 557 connects the substation to the transmission grid.
    net.line.loc[556, 'std_type'] = net.line.loc[801, 'std_type']
	# Change line parameters according to line type
    net.line.c_nf_per_km.loc[556] = net.std_types['line'][net.line.std_type.loc[801]]['c_nf_per_km']
    net.line.r_ohm_per_km.loc[556] = net.std_types['line'][net.line.std_type.loc[801]]['r_ohm_per_km']
    net.line.x_ohm_per_km.loc[556] = net.std_types['line'][net.line.std_type.loc[801]]['x_ohm_per_km']
    net.line.max_i_ka.loc[556] = net.std_types['line'][net.line.std_type.loc[801]]['max_i_ka']

	# ## EHV line 401 (400)
    # Großkrotzenburg to Aschaffenburg. According to Tennet data set, 380 kV 
    # line with 2.6 kA capacity. There is a parallel line. A line between 1884 
    # and 1710 (Staudinger power plant and Großkrotzenburg substation) is added 
    # (with line parameters of EHV 401) and line EHV 412 is removed . To achieve 
    # a similar total capacity between Großkrotzenburg and Aschaffenburg as in 
    # reality, the number of parallel lines is increased to 2. Remptendorf-Redwitz 
    # has 3.6 kA capacity and two parallel systems.
    net.line.parallel.loc[400] = 2
    pp.drop_lines(net, [411])
    new_line_ind = pp.create_line(net, from_bus = 1710, to_bus = 1884, 
                                  length_km = 0.51, std_type = 'LineType_4', 
                                  name = None, index = 411, geodata = None, 
                                  df = 1.0, parallel = 2, in_service = True,
                                  max_loading_percent = 100)
    net.line.loc[new_line_ind, 'name'] = 'EHV Line ' + str(new_line_ind+1)
    net.line.loc[new_line_ind, 'subnet'] = 'EHV1'
    net.line.loc[new_line_ind, 'voltLvl'] = 1

	# ## EHV line 410 (409)
    # Vörden Y-Grohnde 1: According to Tennet 380 kV with 2.6 kA. In the present 
    # data set, however, only 1.3 kA. The subsequent line to Würgassen also has 
    # a capacity of 2.6 kA (EHV 426). For both lines, the line type is changed 
    # from LineType_1 to LineType_2.
    net.line.std_type.loc[[409]] = 'LineType_2'
	# Change line parameters according to line type
    net.line.c_nf_per_km.loc[[409]] = net.std_types['line']['LineType_2']['c_nf_per_km']
    net.line.r_ohm_per_km.loc[[409]] = net.std_types['line']['LineType_2']['r_ohm_per_km']
    net.line.x_ohm_per_km.loc[[409]] = net.std_types['line']['LineType_2']['x_ohm_per_km']
    net.line.max_i_ka.loc[[409]] = net.std_types['line']['LineType_2']['max_i_ka']
    net.line.std_type.loc[[425]] = 'LineType_2'
	# Change line parameters according to line type
    net.line.c_nf_per_km.loc[[425]] = net.std_types['line']['LineType_2']['c_nf_per_km']
    net.line.r_ohm_per_km.loc[[425]] = net.std_types['line']['LineType_2']['r_ohm_per_km']
    net.line.x_ohm_per_km.loc[[425]] = net.std_types['line']['LineType_2']['x_ohm_per_km']
    net.line.max_i_ka.loc[[425]] = net.std_types['line']['LineType_2']['max_i_ka']

	# ## EHV line 193 (192)
    # For orientation: 1474 = 42 = Vieselbach, 316 = Altenfeld, 108 = 1518 Remptendorf.
    # Line Altenfeld-Redwitz is missing! According to 50Hertz grid model EHV 193 
    # (Vieselbach-Altenfeld) has a maximum capacity of 3.6 kA. EHV 222 
    # (Altenfeld-Remptendorf) has a capacity of 2.37 kA. Between Vieselbach-Altenfeld 
    # and Altenfeld-Remptendorf there are 2 line systems each! EHV 192 (191) 
    # (Remptendorf-Redwitz) has a capacity of 3.6 kA and 2 systems!
	# Vieselbach-ALtenfeld
    net.line.loc[192, 'max_i_ka'] = 3.6
    net.line.loc[192, 'parallel'] = 2
	# ALtenfeld-Remptendorf
    net.line.loc[221, 'max_i_ka'] = 2.37
    net.line.loc[221, 'parallel'] = 2
	# Remptendorf-Redwitz
    net.line.loc[191, 'max_i_ka'] = 3.6
    net.line.loc[191, 'parallel'] = 2

	# ## EHV line 261 
    # Oberjettingen-Pulverdingen. According to TransnetBW, the line has a capacity 
    # of 1.6 GW (Simbench 0.4 GW). The line type is changed and a parallel 
    # circuit is added. Oberjettingen-Englstadt (EHV 358) also has a capacity 
    # of 1.6 GW (Simbench 0.8 GW). Therefore: parallel = 2.
    net.line.std_type.loc[260] = 'LineType_1'
	# Change line parameters according to line type
    net.line.c_nf_per_km.loc[260] = net.std_types['line']['LineType_1']['c_nf_per_km']
    net.line.r_ohm_per_km.loc[260] = net.std_types['line']['LineType_1']['r_ohm_per_km']
    net.line.x_ohm_per_km.loc[260] = net.std_types['line']['LineType_1']['x_ohm_per_km']
    net.line.max_i_ka.loc[260] = net.std_types['line']['LineType_1']['max_i_ka']
    net.line.loc[260, 'parallel'] = 2
    net.line.loc[357, 'parallel'] = 2
							  
	# EHV line 558
    # Bus 184: The Weisweiler power plant is connected to the end of the line. 
    # Since the line connects the power plant to the grid, it must be able to 
    # transport the entire power.
    net.line.loc[557, 'parallel'] = 2

	# ## EHV line 599 (598)
    # Knapsack-Sechtem: The sizing of the line seems confusing compared to the 
    # directly connected line in the north, EHV line 487, having a capacity of 
    # 1700 MW (EHV Line 599 has a capacity of 855 MW). The line should be able 
    # to transport 2.6 kA of current, but according to Simbench it only handles 
    # 1.3 kA. Therefore, the line type is changed.
    net.line.std_type.loc[598] = 'LineType_2'
	# Change line parameters according to line type
    net.line.c_nf_per_km.loc[598] = net.std_types['line']['LineType_2']['c_nf_per_km']
    net.line.r_ohm_per_km.loc[598] = net.std_types['line']['LineType_2']['r_ohm_per_km']
    net.line.x_ohm_per_km.loc[598] = net.std_types['line']['LineType_2']['x_ohm_per_km']
    net.line.max_i_ka.loc[598] = net.std_types['line']['LineType_2']['max_i_ka']

	# ## EHV line 643 and EHV line 644 
    # Opladen-Dünnwald: The lines each have 1.3 kA and 220 kV according to Amprion. 
    # According to Simbench, they have 0.65 kA and 220 kV. Therefore, the line 
    # type is changed accordingly.
    pp.change_std_type(net, eid = 642, element = 'line', name = 'LineType_1')
    pp.change_std_type(net, eid = 643, element = 'line', name = 'LineType_1')

	# ## EHV line 662
    # Gremberghoven: The line does not seem to exist in reality. The length is 
    # given as 8 km, whereas the line should be more than 50 km long according 
    # to its connection points. The line is deleted.
    pp.drop_lines(net, [661])

	# ## EHV line 654
    # Utfort-Zensenbusch: Node 3077 is only a junction and not a substation. 
    # Utfort-Zensenbusch and Zensenbusch-Niederrhein have 2.6 kA capacity. 
    # The line type of EHV 654 and EHV 627 is adjusted.
    pp.change_std_type(net, eid = 653, element = 'line', name = 'LineType_2')
    pp.change_std_type(net, eid = 626, element = 'line', name = 'LineType_2')

	# ## EHV line 803 Niederstotzingen-Niederstotzingen Y: 200 m line. The line type 
    # is changed to increase the capacity.
    pp.change_std_type(net, eid = 802, element = 'line', name = 'LineType_1')

	# ## EHV line 12 (11): 
    # Bexbach-Daxlanden (via Otterbach, Lambsheim, Weingarten) should have 
    # 380 kV and 2.6 kA (Simbench 1.3 kA). The line type will be changed. 
    pp.change_std_type(net, eid = 11, element = 'line', name = 'LineType_2')

	# ## EHV line 324 
    # Garenfeld-Eiberg: Should have 1.3 kA. The line type will be changed.
    pp.change_std_type(net, eid = 323, element = 'line', name = 'LineType_1')