#!/usr/bin/env python
# coding: utf-8


"""

======PROJECT======

MathEnergy

======DESCRIPTION======

The functions in this script are used to create timeseries and geoplots.

======Overview functions=====

01: def compare_all_profiles(abs_profiles, freq = 'qh', plot = False)

02: def compare_result_profiles(time_steps, res_variables)

03: def geoplot_mean_line_loading(net, res_variables, 
                                  only_overloaded_lines = False)


​
======AUTHOR, CONTACT & DATE======

Paul Verwiebe, Department of Energy and Resource Management, Technical University of Berlin

verwiebe@tu-berlin.de

June 2021

"""



# ---IMPORT STATEMENTS---

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import seaborn as sns

import pandapower.plotting as pplot

from model.distribution import *
from model.tools import *

import logging
logger = logging.getLogger('mathenergy')


COLORS = {'load' : '#b22222',
          'sgen' : '#2e8b57',
          'gen' : '#8b4513',
          'storage' : '#778899',
          'total' : '#6a5acd'}
		
COLORS_res = {'load' : '#b22222',
              'res_sgen' : '#2e8b57',
              'res_gen' : '#8b4513',
              'res_storage' : '#778899',
              'res_total' : '#6a5acd',
              'res_ext_grid' : 'grey'}



# ---FUNCTIONS---
# 01
def compare_all_profiles(abs_profiles, freq = 'qh', plot = False):
    """Sums all absolute profiles per type in a dataframe.
    
    Parameters
    ----------
    abs_profiles : (dict of pd.DataFrames)
        Dictionary of load, generation and storage profiles.
    
    freq : (str), optional
        'h' for hourly. 'qh' for quarterhourly. The default is 'qh'.
    
    plot : (bool), optional
        Plot all profiles. The default is FALSE.

    Raises
    ------
    ValueError

    Returns
    -------
    profile_sums : (pd.DataFrames)
        Sum of profiles.
        
    Notes
    -----
    - Only returns profile sum when plot is TRUE.
    - Initial profiles are derived from (:obj: 'net').

    """
    if freq == 'qh':
        # Create new dataframe quarterly houred
        profile_sums = pd.DataFrame(index = range(366 * 24 * 4))
    elif freq == 'h':
        # Create new dataframe hourly
        profile_sums = pd.DataFrame(index = range(366 * 24))
    else:
        raise ValueError("This freq is not yet defined. Use 'qh' or 'h'.")

    for feed, df in abs_profiles.items():
        if (feed[1] == 'p_mw' and feed[0]!= 'dcline'):
            if len(df):
                profile_sums = profile_sums.join(df.sum(axis = 1).rename(feed[0]))
    profile_sums['load'] = -profile_sums['load']
    profile_sums['storage'] = -profile_sums['storage']
    # Sum over
    profile_sums['total'] = profile_sums.sum(axis = 1)
    # Set time index
    profile_sums.set_index(get_simbench_timestamps(freq = 'qh'), inplace = True)

    if plot:
        # Plot
        fig, ax = plt.subplots(nrows = 2,  figsize = (15,10))
        colors = [COLORS.get(x, '#333333') for x in profile_sums.columns]
        profile_sums.plot(ax = ax[0], color = colors, alpha = 0.7)
        ax[0].set_title('Demand and Supply Profiles')
        profile_sums.total.plot(ax = ax[1], color = COLORS['total'])
        ax[1].set_title('Balance')
        plt.show()

    return profile_sums


# 02
def compare_result_profiles(time_steps, res_variables):
    """Creates a plot of all result variables: load, gen, sgen, storage, slack etc.
    
    Parameters
    ----------
    time_steps : (range)
        11617 to 14593 for month of may.
    
    res_variables : (dict) 
        dictionary of result variables.

    Returns
    -------
    profile_sums : (pd.DataFrames)
        Sum of profiles.

    """
    profile_sums = pd.DataFrame(index = time_steps)
    index = pd.date_range(start = '2016-05-01 00:00:00', end = '2016-05-31 23:45:00', 
                          freq = '15min')
    
    for feed, df in res_variables.items():
        if feed[1] == 'p_mw':
            if len(df):
                profile_sums = profile_sums.join(df.sum(axis = 1).rename(feed[0]))
        if feed[1] == 'pl_mw':
            if len(df):
                profile_sums = profile_sums.join(df.sum(axis = 1).rename(feed[0]))
                profile_sums[feed[0]] = -profile_sums[feed[0]]
    profile_sums['res_load'] = -profile_sums['res_load']
    profile_sums['res_storage'] = -profile_sums['res_storage']
    # Sum over
    profile_sums['total'] = profile_sums.sum(axis = 1)
    # Set time index
    profile_sums.set_index(index, inplace = True)

    fig, ax = plt.subplots(nrows = 3, figsize = (15,10))
    profile_sums.plot(ax = ax[0], alpha = 1)
    ax[0].set_title('Demand and Supply Profiles')
    ax[0].set_ylabel('MW')
    profile_sums.total.plot(ax = ax[1])
    ax[1].set_title('Balance')
    ax[1].set_ylabel('MW')
    profile_sums.res_ext_grid.plot(ax = ax[2])
    ax[2].set_title('Slack')
    ax[2].set_ylabel('MW')
    plt.show()
	
    return profile_sums


# 03
def geoplot_mean_line_loading(net, res_variables, 
                              only_overloaded_lines = False):
    """creates a geoplot of lines in grid object and assigns colors to lines
	regarding mean line loading.

    Parameters
    ----------
    net : (:obj:pandapowerNet)
        A pandapower grid.
   
    res_variables : (dict)
        dictionary of result variables.
    
    only_overloaded_lines : (bool), optional
        if Ture, only plot 10 lines with highest mean line loading.
		The default is FALSE.

    Returns
    -------
    None.

    """
    net.line['mean_loading_percent'] = res_variables[('res_line', 
                                                      'loading_percent')].mean(axis = 0).to_numpy()
    
    line_cmap_list=[(0, 'green'),
                    (75, 'yellow'),
                    (120, 'red'),
                    (net.line['mean_loading_percent'].max(), 'darkblue')]
    
    dcline_geodata = pd.DataFrame(index = net.dcline.index, columns=['coords'])
    
    for index, row in dcline_geodata.iterrows():
        coords_list = []
        coords_list.append((net.bus_geodata.loc[net.dcline.from_bus[index]]['x'],
                            net.bus_geodata.loc[net.dcline.from_bus[index]]['y']))
        coords_list.append((net.bus_geodata.loc[net.dcline.to_bus[index]]['x'],
                            net.bus_geodata.loc[net.dcline.to_bus[index]]['y']))
        dcline_geodata.iloc[index]['coords'] = coords_list
    
    line_cmap, line_norm = pplot.cmap_continuous(line_cmap_list)
    
    if only_overloaded_lines:
        critical_lines = net.line.mean_loading_percent.nlargest(10)
        lc = pplot.create_line_collection(net, 
                                          critical_lines.index, 
                                          cmap = line_cmap, 
                                          norm = line_norm, 
                                          use_bus_geodata = True, 
                                          z = critical_lines,
                                          linewidth = 2.5,
                                          zorder = 1)
    
    else:
        lc = pplot.create_line_collection(net, 
                                          net.line.index, 
                                          cmap = line_cmap, 
                                          norm = line_norm, 
                                          use_bus_geodata = True, 
                                          z = net.line['mean_loading_percent'],
                                          linewidth = 2.5,
                                          zorder = 1)
        
    bc = pplot.create_bus_collection(net, net.bus.index, zorder = 2, color = 'k', 
                                     size = 0.01)
    dc = pplot.create_line_collection(net, lines = net.dcline.index.to_list(), 
                                      line_geodata = dcline_geodata, 
                                      color = 'black', linewidth = 6, 
                                      alpha = 0.2, zorder = 4)
    sgen_wind_offshore_old = pplot.create_bus_collection(net,
                                                         buses = net.sgen.loc[net.sgen.type == 'wind offshore'].bus.to_list(), 
                                                         size = 0.1, color = 'black', 
                                                         fill = False, zorder = 3)
    
    pplot.draw_collections([lc, bc, sgen_wind_offshore_old, dc])